/*
 *    Copyright (C) 2002  Andrea Marchesini - Bakunin - bakunin@autistici.org
 *                        http://autistici.org/bakunin/
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <sm.h>

int
sm_cond_init (void *cond)
{
  sm_cond_str *tmp;

  if (!cond)
    return SM_ERROR;

  tmp = sm_cond_list;

  while (tmp)
    {
      if (tmp->cond == cond)
	return SM_ERROR;

      tmp = tmp->next;
    }

  if (!(tmp = (sm_cond_str *) malloc (sizeof (sm_cond_str))))
    return SM_MEMORY;

  tmp->cond = cond;
  tmp->w0 = NULL;
  tmp->w1 = NULL;
  tmp->w2 = NULL;

  sm_init (&(tmp->sm));

  sm_lock (&(tmp->sm));

  if (sm_cond_list)
    sm_lock (&(sm_cond_list->sm));

  tmp->next = sm_cond_list;
  sm_cond_list = tmp;

  sm_unlock (&(tmp->sm));

  if (tmp->next)
    sm_unlock (&(tmp->next->sm));

  return SM_OK;
}

int
sm_cond_wait_priority (void *cond, char priority)
{
  sm_cond_str *tmp;
  sm_cond_wait_str *w, *v;

  if (!cond)
    return SM_ERROR;

  tmp = sm_cond_list;

  while (tmp)
    {
      if (tmp->cond == cond)
	break;

      tmp = tmp->next;
    }

  if (!tmp)
    return SM_ERROR;

  sm_lock (&(tmp->sm));

  if (!(w = (sm_cond_wait_str *) malloc (sizeof (sm_cond_wait_str))))
    return SM_MEMORY;

  switch (priority)
    {
    case 1:

      if (!tmp->w0)
	{
	  w->prev = NULL;
	  w->next = NULL;
	  tmp->w0 = w;

	  break;
	}

      v = tmp->w0;
      while (v->next)
	v = v->next;

      w->prev = v;
      w->next = NULL;
      v->next = w;

      break;
    case 0:

      if (!tmp->w1)
	{
	  w->prev = NULL;
	  w->next = NULL;
	  tmp->w1 = w;

	  break;
	}

      v = tmp->w1;
      while (v->next)
	v = v->next;

      w->prev = v;
      w->next = NULL;
      v->next = w;

      break;
    default:

      if (!tmp->w2)
	{
	  w->prev = NULL;
	  w->next = NULL;
	  tmp->w2 = w;

	  break;
	}

      v = tmp->w2;
      while (v->next)
	v = v->next;

      w->prev = v;
      w->next = NULL;
      v->next = w;

    }

  sm_init (&(w->sm));

  sm_lock (&(w->sm));

  sm_unlock (&(tmp->sm));

  if (sm_lock (&(w->sm)))
    sm_cond_signal (cond);

  sm_lock (&(tmp->sm));

  if (w->prev)
    {
      w->prev->next = w->next;
    }
  else
    {
      switch (priority)
	{
	case 1:
	  tmp->w0 = w->next;
	  break;
	case 0:
	  tmp->w1 = w->next;
	  break;
	default:
	  tmp->w2 = w->next;
	  break;
	}
    }

  if (w->next)
    w->next->prev = w->prev;

  free (w);

  sm_unlock (&(tmp->sm));

  return SM_OK;
}

int
sm_cond_signal (void *cond)
{

  sm_cond_str *tmp;

  if (!cond)
    return SM_ERROR;

  tmp = sm_cond_list;

  while (tmp)
    {
      if (tmp->cond == cond)
	break;

      tmp = tmp->next;
    }

  if (!tmp)
    return SM_ERROR;

  if (tmp->w0)
    sm_unlock (&(tmp->w0->sm));
  else if (tmp->w1)
    sm_unlock (&(tmp->w1->sm));
  else if (tmp->w2)
    sm_unlock (&(tmp->w2->sm));

  return SM_OK;
}
