/*
 *    Copyright (C) 2002  Andrea Marchesini - Bakunin - bakunin@autistici.org
 *                        http://autistici.org/bakunin/
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <sm.h>

int
sm_rw_init (sm_rw * x, unsigned int n)
{
  if (!x)
    return SM_ERROR;

  sm_init (&(x->mutex));

  x->rc = 0;

  sm_init (&(x->reader));
  sm_init (&(x->writer));

  x->max_number = n;

  return SM_OK;
}

int
sm_r_lock (sm_rw * x)
{

  if (!x)
    return SM_ERROR;

  if (sm_lock (&(x->reader)))
    return SM_BUSY;

  sm_unlock (&(x->reader));

  if (sm_lock (&(x->mutex)))
    return SM_BUSY;

  while (x->max_number && x->rc >= x->max_number)
    {
      sm_unlock (&(x->mutex));

      sched_yield ();

      if (sm_lock (&(x->mutex)))
	return SM_BUSY;
    }

  x->rc++;

  if (x->rc == 1)
    if (sm_lock (&(x->writer)))
      return SM_BUSY;

  sm_unlock (&(x->mutex));

  return SM_OK;
}

int
sm_r_unlock (sm_rw * x)
{
  if (!x)
    return SM_ERROR;

  if (sm_lock (&(x->mutex)))
    return SM_BUSY;

  x->rc--;
  if (x->rc < 0)
    x->rc = 0;

  if (!x->rc)
    sm_unlock (&(x->writer));

  sm_unlock (&(x->mutex));

  return SM_OK;
}

int
sm_w_lock (sm_rw * x)
{
  if (!x)
    return SM_ERROR;

  if (sm_lock (&(x->reader)))
    return SM_BUSY;
  if (sm_lock (&(x->writer)))
    return SM_BUSY;

  return SM_OK;
}

int
sm_w_unlock (sm_rw * x)
{

  if (!x)
    return SM_ERROR;

  sm_unlock (&(x->writer));
  sm_unlock (&(x->reader));

  return SM_OK;
}
